
def called_by(*source_methods):
    """Decorator to mark either the calling function or the originator of a message.
    Can accept multiple source methods.

    ### Usage:
    @called_by(ClassName.method, some_func_foo, otherclass.foo)
    
    This achieves two goals:
    1) Allows for jump-clicking to the source of a message or a calling function.
    2) Validates that the 'source' method exists."""

    def decorator(func):
        return func
    return decorator

    # NOTE: There's no need for any validation here. If a source method is not found,
    # Python will raise an AttributeError. That is the validation.



"""Tic-Tac-Toe game script for TextualGames"""

# Textual imports
from textual.app import on
from textual.containers import Container, Horizontal
from textual.widgets import Button

# TextualGames imports
from textual_games.game import GameBase
from textual_games.grid import Grid
from textual_games.enums import PlayerState

# NOTE: whitespace is important here:
x_token = r"""
 \_/ 
 / \ """
o_token = r"""  __ 
 /  \
 \__/"""

def loader():
    """Required function that returns the game's main widget"""
    return TicTacToe

        
class TicTacToe(GameBase):

    game_name = "Tic-Tac-Toe"

    def compose(self):

        self.grid_size = 3
        self.grid = Grid(
            rows=3,
            columns=3,
            grid_width=34,          # NOTE: The sizes and formatting can be tricky to get perfect, so I've
            grid_height=18,         # decided its easiest to make each game enter the grid size and cell size manually.
            grid_gutter=0,              # If writing a game, you'll need to just play with these numbers until 
            player1_token=x_token,      # everything looks exactly the way you'd like.
            player2_token=o_token,      
            cell_size=5,            # <- nice big cells for Tic-Tac-Toe
            classes="grid onefr"
        )

        with Container(id="content", classes="onefr centered"):
            yield self.grid
        with Horizontal(classes="centered wide footer"):
            yield Button("Restart", id="restart", classes="centered")

    async def on_mount(self):
        self.restart() 

    @on(Grid.RestartGame)
    @on(Button.Pressed, "#restart")
    def restart(self):
        self.grid.restart_grid()
        self.post_message(self.StartGame(
            game = self,
            rows = self.grid_size,
            columns = self.grid_size,
            max_depth = 9
        ))

    #* Called by: calculate_winner in TextualGames class.
    def calculate_winner(self, board: list[list[int]]) -> PlayerState | None:
        """Returns a PlayerState if game is over, else returns None."""

        rows      = board
        columns   = list(zip(*board))
        main_diag = [[board[i][i] for i in range(self.grid_size)]]
        anti_diag = [[board[i][self.grid_size - i - 1] for i in range(self.grid_size)]]

        # Combine all possible lines into a single list
        lines = (rows + columns + main_diag + anti_diag)

        def check_line(line, player):
            return all(cell == player for cell in line)
        
        for player in (1, 2):
            if any(check_line(line, player) for line in lines):     # Check for winner
                return PlayerState.PLAYER1 if player == 1 else PlayerState.PLAYER2

        if all(cell != 0 for row in board for cell in row):         # Check for draw
            return PlayerState.EMPTY
        
        return None        # game not over yet


from __future__ import annotations
from copy import deepcopy
from asyncio import sleep

# Textual imports
from rich.text import Text
from textual import work, on
from textual.worker import Worker, WorkerState
from textual.message import Message
from textual.widget import Widget

# TextualGames imports
from textual_games.enums import PlayerState, GridGravity
from textual_games.game import GameBase

class GameManager(Widget):

    class ChangeTurn(Message):
        """Posted by `start_game`, `cell_pressed`, and `computer_turn_orch` in GameManager. \n
        Handled by `change_turn` in TextualGames."""
        def __init__(self, value: PlayerState):
            super().__init__()
            self.value = value

    class ComputerMove(Message):
        """Posted by `computer_turn_orch` when the AI has chosen a move. \n
        Handled by: `play_computer_move` in TextualGames class."""
        def __init__(self, row: int, col: int):
            super().__init__()
            self.row = row
            self.col = col

    class GameOver(Message):
        """Posted by: `cell_pressed` and `computer_turn_orch` in GameManager. \n
        Handled by `game_over` in TextualGames."""
        def __init__(self, result: PlayerState):
            super().__init__()
            self.result = result


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.display = False
        self.game_running = False
    
    #* Called by: `mount_grid` and `restart` in TicTacToe.
    def start_game(self, event: GameBase.StartGame):

        self.game_running = True
        self.game = event.game       #! This is not currently used, but could be useful for future features.
        self.rows = event.rows          # right now things are mostly decoupled.
        self.columns = event.columns
        self.max_depth = event.max_depth
        self.gravity = event.gravity
        self.empty_cells = self.rows*self.columns
        self.int_board = [[0 for _ in range(self.columns)] for _ in range(self.rows)]
        self.move_counter = 0

        self.log(
            "Starting game with settings:\n"
            f"Rows: {self.rows}\n"
            f"Columns: {self.columns}\n"
            f"Max depth: {self.max_depth}\n"
            f"Gravity: {self.gravity}\n"
        )

        self.post_message(self.ChangeTurn(PlayerState.PLAYER1))
        self.notify("Game started", timeout=1.5)

    def end_game(self, game_result: PlayerState):
        self.game_running = False
        self.post_message(self.GameOver(game_result))

    #* Called by: cell_pressed in Grid class
    async def cell_pressed(self, row: int, column: int):
        self.log.info(f"manager Cell pressed: {row}, {column}")

        self.log.info(f"self.game_running: {self.game_running}")

        if self.int_board[row][column] != 0:
            self.notify("Cell already taken", timeout=1)
            return
        if not self.game_running:
            return

        self.int_board[row][column] = 1
        self.move_counter += 1
        self.empty_cells -= 1
        game_result = self.app.calculate_winner(self.int_board)
        if game_result is not None:
            self.end_game(game_result)
            return
    
        self.post_message(self.ChangeTurn(PlayerState.PLAYER2))     # Change turn to AI
        # NOTE: I let the main app call the computer_turn_orch method to ensure the
        # UI is updated before it starts thinking.

    #* Called by: change_turn in main App.
    async def computer_turn_orch(self):

        self.minimax_counter = 0
        self.pruning_counter = 0
        self.depth_limit_counter = 0

        board_copy = deepcopy(self.int_board)
        worker = self.computer_turn_worker(board_copy)
        ai_row, ai_col = await worker.wait()
        if ai_row is None:
            raise ValueError("AI made an invalid move.")

        self.log(
            f"Minimax counter: {self.minimax_counter}\n"
            f"Branches pruned: {self.pruning_counter}\n"
            f"Times depth limit reached: {self.depth_limit_counter}\n"
        )        

        self.int_board[ai_row][ai_col] = 2              # Apply AI move to integer board
        self.move_counter += 1
        self.empty_cells -= 1
        self.post_message(self.ComputerMove(ai_row, ai_col))  # Updates the cell state in the Grid

        game_result = self.app.calculate_winner(self.int_board)
        if game_result is not None:
            self.end_game(game_result)
            return
        
        self.post_message(self.ChangeTurn(PlayerState.PLAYER1))     # Change turn back to human


    def get_possible_moves(self, board: list[list[int]]) -> list[tuple[int, int]]:

        possible_moves = []
        
        if self.gravity == GridGravity.DOWN:
            for col in range(self.columns):
                for row in range(self.rows):
                    if row != self.rows-1:                      # if not last row:
                        if board[row+1][col] == 0:         # is row below empty?
                            continue
                        else:
                            if board[row][col] == 0:
                                possible_moves.append((row, col))
                            break
                    else:                                    
                        possible_moves.append((row, col))       # it should only get here if column is empty
        else:
            for row in range(self.rows):
                for col in range(self.columns):
                    if board[row][col] == 0:
                        possible_moves.append((row, col))


        return possible_moves


    @work(thread=True, exit_on_error=False)
    async def computer_turn_worker(self, board: list[list[int]]) -> tuple[int, int]:

        await sleep(0.5)            # Artificial delay to simulate thinking time
        _, best_move = self.minimax(
            board,
            depth=0,                 # always start at depth 0
            is_maximizing=True,      # AI is maximizer
            alpha=float('-inf'),
            beta=float('inf'),
        )   
        return best_move

    def minimax(
        self,
        board: list[list[int]],
        depth: int,
        is_maximizing: bool,
        alpha: float,
        beta: float,
    ) -> tuple[int, tuple[int, int]]:
        """ | Arg           | Description 
            |---------------|---------------------
            | board         | - The current game board state
            | depth         | - Current depth in the game tree
            | is_maximizing | - True if AI's turn (maximizing), False if human's turn (minimizing)

            Returns:
                tuple[int, tuple[int, int]]: best_score, best_move (as tuple of coordinates)"""

        self.minimax_counter += 1

        result = self.app.calculate_winner(board)

        # Base cases: game over scenarios
        if result == PlayerState.PLAYER1:     # Human is minimizer
            return -10 + depth, None
        elif result == PlayerState.PLAYER2:   # AI is maximizer
            return 10 - depth, None
        elif result == PlayerState.EMPTY:     # Draw
            return 0, None

        if depth >= self.max_depth:
            self.depth_limit_counter += 1
            return 0, None
    
        best_move  = (None, None)
        best_score = float('-inf') if is_maximizing else float('inf')
        player     =             2 if is_maximizing else 1

        possible_moves = self.get_possible_moves(board)

        for move in possible_moves:
            row, col = move

            board[row][col] = player
            score, _, = self.minimax(board, depth + 1, not is_maximizing, alpha, beta)
            board[row][col] = 0         # Undo move

            if is_maximizing and score > best_score:
                    best_score = score
                    alpha = max(score, alpha)
                    best_move = (row, col)
            elif not is_maximizing and score < best_score:
                    best_score = score
                    beta = min(score, beta)
                    best_move = (row, col)

            if beta <= alpha:
                self.pruning_counter += 1
                break

        return best_score, best_move

    @on(Worker.StateChanged)
    def worker_state_changed(self, event: Worker.StateChanged) -> None:
        if event.state == WorkerState.SUCCESS:
            self.log(Text(f"Worker {event.worker.name} completed successfully", style="green"))




from __future__ import annotations

# from rich.text import Text
from textual import on
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Static
from textual.message import Message
from textual.binding import Binding

from textual_games.source_decorator import called_by
from textual_games.enums import PlayerState, GridFocusType, GridGravity

class Cell(Static):

    state = reactive(PlayerState.EMPTY)
    """This is modified by `cell_pressed` method in the GameManager class
    whenever a move is chosen by either the player or the computer."""

    class Pressed(Message):
        """Posted by: `on_click` in Cell. \n
        Handled by: `cell_pressed` in TextualGames."""
        def __init__(self, cell: Cell):
            super().__init__()
            self.cell = cell

    class HoverEnter(Message):
        def __init__(self, cell: Cell) -> None:
            self.cell = cell
            """The cell that was entered (hovered over)."""
            super().__init__()

    def __init__(
            self,
            row: int,
            column: int,
            cell_size: int,
            name: str | None = None,
            id: str | None = None,
            classes: str | None = None,
        ):
        """ | Arg     | Description 
            |---------|-------------
            | row     | - The row of the cell
            | column  | - The column of the cell
            | name    | - The name of the widget
            | id      | - The ID of the widget in the DOM
            | classes | - The CSS classes for the widget """
            
        super().__init__(name=name, id=id, classes=classes)
        self.row = row
        self.column = column
        self.styles.width = (cell_size * 2)+1  # width of double the height makes a square (roughly) in monospace
        self.styles.height = cell_size

    def render(self):

        if self.state == PlayerState.PLAYER1:
            return self.parent.player1_token
        elif self.state == PlayerState.PLAYER2:
            return self.parent.player2_token
        else:
            return ""
        
    def watch_state(self, value):

        if self.parent.player1_color and self.parent.player2_color:
            if value == PlayerState.PLAYER1:
                self.set_class(True, 'red')
            elif value == PlayerState.PLAYER2:
                self.set_class(True, 'yellow')
            else:
                self.set_class(False, 'red', 'yellow')

    def watch_mouse_hover(self, value: bool) -> None:
        """OVERRIDE: Update from CSS if mouse over state changes.
        Textual addition: posts HoverEnter message."""

        if self._has_hover_style:
            self._update_styles()
        if value:
            self.post_message(self.HoverEnter(self))

    def on_click(self):
        self.post_message(self.Pressed(self))


class Grid(Widget):

    #* Handled by: cell_chosen in TextualGames class.
    class CellChosen(Message):
        def __init__(self, row: int, column: int):
            super().__init__()
            self.row = row
            self.column = column

    #* Sent by: action_restart in Grid class.
    #* Handled by: restart in the Game class that uses the grid.
    class RestartGame(Message):
        pass

    BINDINGS = [
        Binding("left", "left", "Move left"),
        Binding("right", "right", "Move right"),
        Binding("up", "up", "Move up"),
        Binding("down", "down", "Move down"),
        Binding("enter", "select", "Select"),
        Binding("r", "restart", "Restart"),
    ]

    # focus_grid: reactive[list[list[int]]] = reactive([])        # 2d list
    focus_string: str = reactive("")

    def __init__(
            self,
            rows: int,
            columns: int,
            grid_width: int,
            grid_height: int,
            cell_size: int,
            grid_gutter: int = 1,
            player1_token: str = '',
            player2_token: str = '',
            player1_color: str | None = None,
            player2_color: str | None = None,
            focus_mode: GridFocusType = GridFocusType.ALL,
            gravity: GridGravity = GridGravity.NONE,
            name: str | None = None,
            id: str | None = None,
            classes: str | None = None,
        ):  #! TODO update docstring VVV
        """ | Arg             | Description 
            |-----------------|-------------
            | rows            | - The number of rows in the grid
            | columns         | - The number of columns in the grid
            | grid_width      | - The width of the grid
            | grid_height     | - The height of the grid
            | cell_size       | - The size of each cell
            | grid_gutter     | - The space between cells
            | player1_token   | - The token for player 1
            | player2_token   | - The token for player 2
            | player1_color   | - The color of player 1's token
            | player2_color   | - The color of player 2's token
            | focus_mode      | - The direction of focus
            | gravity         | - The direction of gravity
            | name            | - The name of the widget
            | id              | - The ID of the widget in the DOM
            | classes         | - The CSS classes for the widget """
        
        super().__init__(name=name, id=id, classes=classes)
        self.rows:      int = rows
        self.columns:   int = columns
        self.cell_size: int = cell_size
        self.player1_token: str = player1_token
        self.player2_token: str = player2_token
        self.player1_color: str = player1_color
        self.player2_color: str = player2_color
        self.focus_mode: GridFocusType = focus_mode
        self.gravity : GridGravity = gravity

        self.styles.grid_size_rows = self.rows
        self.styles.grid_size_columns = self.columns
        self.styles.grid_gutter_vertical = grid_gutter
        self.styles.grid_gutter_horizontal = grid_gutter
        self.styles.width = grid_width
        self.styles.height = grid_height

        #! CHANGE FOCUS GRID TO REACTIVE. MODEL/VIEW PATTERN.

    def compose(self):

        for row in range(self.rows):
            for col in range(self.columns):
                yield Cell(
                    row = row,
                    column = col,
                    cell_size = self.cell_size,
                    id = f"cell_{row}_{col}",
                    classes = "gridcell bordered centered"
                )

    #* Called by: game_over in TextualGames class.
    def clear_focus(self):
        "Disables focus on the grid and removes the focusing class from all cells."

        self.can_focus = False
        for cell in self.query_children(Cell):
            cell.remove_class("focusing")
        self.focus_string = "0" * (self.rows * self.columns)

    #* Called by: restart in TextualGames class. Used by games to both start and restart.
    def restart_grid(self):
        "Clears the grid, re-enables focus, and focuses cell (0,0)"
        self.log.debug("Restarting grid...")

        for cell in self.query_children(Cell):
            cell.state = PlayerState.EMPTY
        self.focus_string = "0" * (self.rows * self.columns)
        self.can_focus = True
        self.focus()
        self.focus_cell(0, 0) 

    def check_action(
        self, action: str, parameters: tuple[object, ...]
    ) -> bool | None:  
        """Check if an action may run."""
        if action == "up" or action == "down":
            if self.focus_mode == GridFocusType.COLUMNS:
                return False
        return True

    @on(Cell.Pressed)
    def action_select(self):
        self.log.debug("Action: Select")
        row_index, col_index = self.string_to_coordinates(self.focus_string)
        self.post_message(self.CellChosen(row_index, col_index))
        self.query_one(f"#cell_{row_index}_{col_index}").state = PlayerState.PLAYER1

    def action_app_focus(self):
        self.app.action_focus_next()

    def action_restart(self):
        self.log.debug("Action: Restart")
        self.post_message(self.RestartGame())


    ###~ Grid Focusing System ~###

    #* Movement:
    # --------------

    @on(Cell.HoverEnter)
    def cell_hovered(self, event: Cell.HoverEnter):
        self.log.debug(f"Cell hovered: {event.cell.row}, {event.cell.column}")
        if not self.can_focus:
            return
        self.focus_cell(event.cell.row, event.cell.column)   

    def action_left(self):
        self.log.debug("Action: Left")
        self.refresh_bindings()
        row_index, col_index = self.string_to_coordinates(self.focus_string)
        self.focus_cell(
            row_index, (col_index-1 if col_index != 0 else self.columns-1)
        ) # subtract 1 if not first column, else go to last column

    def action_right(self):
        self.log.debug("Action: Right")
        self.refresh_bindings()
        row_index, col_index = self.string_to_coordinates(self.focus_string)
        self.focus_cell(
            row_index, (col_index+1 if col_index != self.columns-1 else 0)
        ) # add 1 if not last column, else go to first column

    def action_up(self):
        self.log.debug("Action: Up")
        self.refresh_bindings()
        row_index, col_index = self.string_to_coordinates(self.focus_string)
        self.focus_cell(
            (row_index-1 if row_index != 0 else self.rows-1), col_index
        ) # subtract 1 if not first row, else go to last row

    def action_down(self):
        self.log.debug("Action: Down")
        self.refresh_bindings()
        row_index, col_index = self.string_to_coordinates(self.focus_string)
        self.focus_cell(
            (row_index+1 if row_index != self.rows-1 else 0), col_index
        ) # add 1 if not last row, else go to first row

    #* Logic:
    # --------------

    def watch_focus_string(self, old: str, new: str):
        """Using the model/view pattern, this method watches the focus_grid and updates the UI accordingly."""

        old_row, old_col = self.string_to_coordinates(old)
        new_row, new_col = self.string_to_coordinates(new)

        self.log(
            f"Old string: {old} \n"
            f"old_row: {old_row}, old_col: {old_col} \n"
            f"New string: {new} \n"
            f"new_row: {new_row}, new_col: {new_col}"
        )

        if old_row is not None and old_col is not None:
            self.query_one(f"#cell_{old_row}_{old_col}").remove_class("focusing")
        if new_row is not None and new_col is not None:
            self.query_one(f"#cell_{new_row}_{new_col}").add_class("focusing")


    @called_by(restart_grid, cell_hovered,
        action_left, action_right, action_up, action_down)
    def focus_cell(self, row: int, col: int):

        if self.gravity == GridGravity.DOWN:
            for newrow in range(self.rows):
                if newrow != self.rows-1:                      # if not last row
                    self.log("Not the last row...")
                    if self.query_one(f"#cell_{newrow+1}_{col}").state == PlayerState.EMPTY:
                        self.log("Cell below is empty. Continuing...")
                        continue
                    else:
                        self.log(f"Cell below is not empty. Focusing cell {newrow}, {col}")
                        row = newrow
                        break
                else:                   # if it is last row
                    self.log(f"Last row. Focusing cell {newrow}, {col}")
                    row = newrow
                    break

        self.focus_string = self.coordinates_to_string(row, col)


    @called_by(action_select, watch_focus_string,
        action_left, action_right, action_up, action_down)
    def string_to_coordinates(self, string: str) -> tuple[int, int] | tuple[None, None]:

        if '1' in string:
            index = string.index('1')  # Find the index of the '1' in the string
            row = index // self.columns    # Convert index to row
            col = index % self.columns     # Convert index to column
            return row, col
        else:
            return None, None
        
    @called_by(focus_cell)
    def coordinates_to_string(self, row: int, col: int) -> str:

        new_focus = [0] * (self.rows * self.columns)
        new_focus[(row * self.columns) + col] = 1
        return "".join(map(str, new_focus)) 










from enum import Enum

class PlayerState(Enum):
    """The state of a cell in the game. \n
    Can be EMPTY, PLAYER1, or PLAYER2."""
    EMPTY = 0
    PLAYER1 = 1
    PLAYER2 = 2

class GridFocusType(Enum):
    ALL = "all"
    ROWS = "rows"
    COLUMNS = "columns"

class GridGravity(Enum):
    UP = "up"
    DOWN = "down"
    LEFT = "left"
    RIGHT = "right"
    NONE = "none"


"""Base class for all games. \n"""

# from abc import ABC, abstractmethod
from __future__ import annotations

from textual.message import Message
from textual.widget import Widget

from textual_games.enums import GridGravity

# from textual_games.enums import PlayerState


class GameBase(Widget):

    class StartGame(Message):
        """Posted when a game is either mounted or restarted. \n
        Handled by start_game in TextualGames class."""

        def __init__(
                self,
                game: GameBase,
                rows: int,
                columns: int,
                max_depth: int,
                gravity: GridGravity = GridGravity.NONE
            ):
            super().__init__()
            self.game = game
            self.rows = rows
            self.columns = columns
            self.max_depth = max_depth
            self.gravity = gravity

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            self.calculate_winner
        except AttributeError:
            raise NotImplementedError("Game must implement calculate_winner method.")
        
        try:
            self.game_name
        except AttributeError:
            raise NotImplementedError("Game must implement game_name attribute.")
        
        try:
            self.restart
        except AttributeError:
            raise NotImplementedError("Game must implement restart method.")

    # #* Called by: calculate_winner in Main App.
    # def calculate_winner(self, board: list[list[int]]) -> PlayerState | None:
    #     """Returns a PlayerState if game is over, else returns None."""
    #     pass


# Standard library imports
from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from textual.events import Click

from rich.console import RenderableType
from rich.text import Text

from textual.widgets import Static
from textual.message import Message
from textual.binding import Binding



class SimpleButton(Static):
    """A simple button widget. Note that this does not inherit from Button, but from Static.

    This is designed to copy the basic functionality of the Textual button, but in a
    more minimalistic way that changes some of the core functionality and appearance.

    It can be any size (down to 1x1). It copies the `Pressed` event from normal buttons.
    It also renders itself using the rich Text class, so you can set
    the overflow and no_wrap properties.

    By default, no_wrap is `True` and overflow is `crop`. This meas the button
    text will not wrap and will be cropped if it is too long. You can change this by using
    the 'no_wrap' and 'overflow' arguments in the constructor.

    It also contains the `HoverEnter` and `HoverLeave` events, which are triggered when the
    mouse enters or leaves the button. These provide convenient messages for hovering over the button,
    so that the hover event can be handled by the parent widget.

    EXAMPLE USAGE:
    - yield button:
    ```
    yield SimpleButton("your_button_text", id="sample_id")
    ```
    - In some cases you'll want word wrap enabled, without overflow:
    ```
    yield SimpleButton("your_button_text", no_wrap=None, overflow=None)
    ```
    - message handler:
    ```
    @on(SimpleButton.Pressed)   
    def on_button_pressed(self, message: SimpleButton.Pressed) -> None:   
        print("Button was pressed!")
    ```
    """

    BINDINGS = [Binding("enter", "press", "Press button", show=False)]

    ### MESSAGES ###

    class Pressed(Message):
        """Event sent when a `SimpleButton` is pressed.   

        Can be handled using `on_simple_button_pressed` (in a subclass of
        [`Button`][textual.widgets.Button] or in a parent widget in the DOM.)   
        OR   
        by using @on(SimpleButton.Pressed)"""

        def __init__(self, button: SimpleButton) -> None:
            super().__init__()

            self.button: SimpleButton = button
            """The button that was pressed.   
            Note that it is a SimpleButton object.   
            You can access the button's properties:
            - button.id
            - button.name
            - button.index   (if you set an index)
            - etc"""

        @property
        def control(self) -> SimpleButton:
            """This is required to be able to use the 'selector' property
            when using the message handler."""

            return self.button
        
    class HoverEnter(Message):
        def __init__(self, button: SimpleButton) -> None:
            self.button: SimpleButton = button
            """The button that was entered (hovered over)."""
            super().__init__()

        @property
        def control(self) -> SimpleButton:
            return self.button

    class HoverLeave(Message):
        def __init__(self, button: SimpleButton) -> None:
            self.button: SimpleButton = button
            """The button that was left."""
            super().__init__()

        @property
        def control(self) -> SimpleButton:
            return self.button

    ### END MESSAGES ###
    ###  SimpeButton MAIN  ###

    def __init__(
            self,
            *args,
            justify:  str | None = None,
            overflow: str | None = "crop",
            no_wrap: bool | None = True,
            index:    int | None = None,
            **kwargs
        ) -> None:
        """
        Custom constructor for SimpleButton. Adds 'justify', 'overflow', and 'no_wrap' arguments for
        the Rich Text class. args and kwargs are passed to the Static constructor.

        Args:
            renderable: A Rich renderable, or string containing console markup.
            justify (str, optional): Justify method: "left", "center", "full", "right". Defaults to None.
            overflow (str, optional): Overflow method: "crop", "fold", "ellipsis". Defaults to None.
            no_wrap (bool, optional): Disable text wrapping, or None for default. Defaults to None.
            index (int, optional): Allows setting index for the button. This is useful if the button
                is part of a list of buttons. Defaults to None.
            expand: Expand content if required to fill container.
            shrink: Shrink content if required to fill container.
            markup: True if markup should be parsed and rendered.
            name: Name of widget.
            id: ID of Widget.
            classes: Space separated list of class names.
            disabled: Whether the static is disabled or not."""

        super().__init__(*args, **kwargs)
        self.can_focus = True
        self.no_wrap = no_wrap
        self.overflow = overflow
        self.justify = justify
        self.index = index

    def render(self) -> RenderableType:
        """Pass the options from the constructor into the Rich Text class."""
        return Text(
            text=str(self.renderable),
            justify=self.justify,
            no_wrap=self.no_wrap,
            overflow=self.overflow
            )
    
    def watch_mouse_hover(self, value: bool) -> None:
        """OVERRIDE: Update from CSS if mouse over state changes.
        Textual addition: posts HoverEnter / HoverLeave messages."""

        if self._has_hover_style:
            self._update_styles()
        if value:
            self.post_message(self.HoverEnter(self))
        else:
            self.post_message(self.HoverLeave(self))
        
    def on_click(self, event: Click) -> None:
        """Called when the button is clicked. Posts a message 'Pressed'.
        Use the message handler to handle the event:   
        ```
        @on(SimpleButton.Pressed)   
        def on_button_pressed(self, message: SimpleButton.Pressed) -> None:   
            print("Button was pressed!")
        ``` """
        self.post_message(self.Pressed(self))

    def action_press(self) -> None:
        self.post_message(self.Pressed(self))


"""Connect Four game script for TextualGames"""

from __future__ import annotations

# Textual imports
from textual.app import on
from textual.containers import Container, Horizontal
from textual.widgets import Button

from textual_games.game import GameBase
from textual_games.grid import Grid, GridFocusType, GridGravity
from textual_games.enums import PlayerState


def loader():
    """Required function that returns the game's main widget"""
    return ConnectFour
        

class ConnectFour(GameBase):

    game_name = "Connect Four"

    def compose(self):

        focus_mode = GridFocusType.COLUMNS      # This BS is necessary because intellisense wouldn't work
        gravity = GridGravity.DOWN              # properly if I tried inserting it down there directly.

        self.grid = Grid(
            rows=6,
            columns=7,
            grid_width=50,
            grid_height=19,
            grid_gutter=0, 
            player1_color="red",
            player2_color="yellow",
            cell_size=3,
            focus_mode=focus_mode,
            gravity=gravity,
            classes="grid onefr"
        )        

        with Container(id="content", classes="onefr centered"):
            yield self.grid
        with Horizontal(classes="centered wide footer"):
            yield Button("Restart", id="restart", classes="centered")

    async def on_mount(self):
        self.restart()

    @on(Button.Pressed, "#restart")
    def restart(self):
        self.grid.restart_grid()
        self.post_message(self.StartGame(
            game = self,
            rows = 6,
            columns = 7,
            max_depth = 7,
            gravity=GridGravity.DOWN
        ))

    #* Called by: calculate_winner in TextualGames class.
    def calculate_winner(self, board: list[list[int]]) -> PlayerState | None:
        """Returns a PlayerState if the game is over, else returns None."""

        rows = len(board)
        cols = len(board[0])
        streak_to_win = 4

        lines = []                           # Pre-calculate all possible lines
        lines.extend(board)                                                                 # Add all rows
        lines.extend([[board[row][col] for row in range(rows)] for col in range(cols)])     # Add all columns

        for row in range(rows):              # Add all diagonals (main and anti)
            for col in range(cols):

                # Main diagonal (top-left to bottom-right)
                if row <= rows - streak_to_win and col <= cols - streak_to_win:
                    lines.append([board[row + i][col + i] for i in range(streak_to_win)])
                
                # Anti-diagonal (top-right to bottom-left)
                if row <= rows - streak_to_win and col >= streak_to_win - 1:
                    lines.append([board[row + i][col - i] for i in range(streak_to_win)])

        def check_line(line: list[int], player: int) -> bool:
            """Checks if a line contains 4 consecutive pieces of the same player."""
            count = 0
            for cell in line:
                count = count + 1 if cell == player else 0
                if count == streak_to_win:
                    return True
            return False
    
        for player in (1, 2):
            if any(check_line(line, player) for line in lines):     # Check for winner
                return PlayerState.PLAYER1 if player == 1 else PlayerState.PLAYER2

        if all(cell != 0 for row in board for cell in row):         # Check for draw
            return PlayerState.EMPTY

        return None         # game not over yet




# from __future__ import annotations
import importlib.util
import os
from typing import Dict, Any, Type, List

# 3rd party imports
from textual_pyfiglet import FigletWidget

# Textual imports
from textual import on, work
from textual.binding import Binding
from textual.app import App, SystemCommand
from textual.screen import Screen
from textual.message import Message
from textual.containers import Container, Horizontal
from textual.widgets import Footer, Label, Static
from textual.dom import DOMNode
from textual.widget import Widget

# TextualGames imports
import textual_games.games as games
from textual_games.source_decorator import called_by
from textual_games.spinner import SpinnerWidget, ScrollingLine
from textual_games.simplebutton import SimpleButton
from textual_games.manager import GameManager
from textual_games.enums import PlayerState
from textual_games.grid import Grid
from textual_games.game import GameBase


class GameEntry(Widget):

    class GameSelected(Message):
        """Posted by: entry_pressed in GameEntry. \n
        Handled by: game_selected in TextualGames."""
        def __init__(self, game_class: Type[Any]):
            super().__init__()
            self.game_class = game_class

    def __init__(self, game_name: str, game_class: Type[Any], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.game_name = game_name
        self.game_class = game_class

    def compose(self):
        yield SimpleButton(self.game_name, classes="wide centered")

    @on(SimpleButton.Pressed)
    def entry_pressed(self, event: SimpleButton.Pressed):
        self.log.info(f"Pressed: {self.game_name}")
        self.post_message(self.GameSelected(self.game_class))

    def focus(self):
        self.query_one(SimpleButton).focus()


class TextualGames(App):

    CSS_PATH = ["css.tcss"]

    BINDINGS = [
        # Binding("escape", "foo", "Menu", show=True),
        Binding("up", "menu_previous", "Previous"),
        Binding("down", "menu_next", "Next"),
    ]

    # COMMAND_PALETTE_BINDING = "escape"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def compose(self):

        self.game_manager = GameManager()
        yield self.game_manager

        with Horizontal(id="header", classes="wide header tall"):
            yield FigletWidget("Textual Games", font="small_slant")
        with Horizontal(id="animation_header", classes="h1 wide"):
            yield ScrollingLine(classes="wide")
        with Horizontal(id="turn_header", classes="wide header turnlabel"):
            yield Label(id="turn_label", classes="auto centered")
            yield SpinnerWidget("line", id="spinner", classes="auto centered")
        with Container(id="content", classes="onefr centered"):
            yield Static()
        yield Footer()

    def get_system_commands(self, screen: Screen):
        yield from super().get_system_commands(screen)  
        yield SystemCommand("Main Menu", "Exit game and go back to main menu", self.mount_games_menu)

    def on_mount(self):
        self.call_after_refresh(self.load_games)

    #* Called by: on_mount, directly above.
    async def load_games(self):

        self.content_window = self.query_one("#content")
        self.loader = GameLoader.with_default_games()            # Load only the package's built-in games
        # loader = GameLoader.with_default_games(["/path/to/user/games"])    # Load package and user games
        # loader = GameLoader("/path/to/games", "/path/to/more/games")       # specify directories manually

        games = self.loader.discover_games()
        self.log(games)

        self.all_games = [
            GameEntry(key, value, classes="wide centered") for key, value in games.items()
        ]
        await self.mount_games_menu()

    async def mount_games_menu(self):

        self.game_manager.game_running = False
        await self.content_window.remove_children()
        await self.content_window.mount_all(self.all_games)
        self.query_one("#turn_label").update("")

        self.query_one("#header").display = True
        self.query_one("#animation_header").display = True
        self.query_one("#spinner").visible = False
        self.query_one("#turn_header").display = False
        
        self.content_window.query_one(GameEntry).focus()

    def action_menu_next(self) -> None:
        self.refresh_bindings()
        self.screen.focus_next()

    def action_menu_previous(self) -> None:
        self.refresh_bindings()
        self.screen.focus_previous()

    def check_action(
        self, action: str, parameters: tuple[object, ...]
    ) -> bool | None:  
        """Check if an action may run."""
        if action == "menu_next" or action == "menu_previous":
            if self.game_manager.game_running:
                return False
        return True
    
    ###~ Event Handlers ~###

    @called_by(GameEntry.entry_pressed)
    @on(GameEntry.GameSelected)
    def game_selected(self, event: GameEntry.GameSelected):

        self.log.info(f"game_selected in main app received: {event.game_class}")

        self.content_window.remove_children()
        self.current_game = event.game_class()
        self.content_window.mount(self.current_game)

    #* Called by on_mount or restart in a game widget.
    @on(GameBase.StartGame)
    def start_game(self, event: GameBase.StartGame):
        self.workers.cancel_all()
        self.query_one("#header").display = False
        self.query_one("#animation_header").display = False
        self.query_one("#turn_header").display = True
        self.game_manager.start_game(event)
        self.call_after_refresh(self.query_one(Footer).refresh, recompose=True)


    @called_by(
        GameManager.cell_pressed,
        GameManager.computer_turn_orch,
        GameManager.minimax
    )
    def calculate_winner(self, board: List[List[int]]) -> PlayerState | None:
        """Returns a PlayerState if game is over, else returns None. \n
        Playerstate can be PLAYER1, PLAYER2, or EMPTY. (Empty here means draw/tie game)"""
        # NOTE: This does not use an event because we need it to wait for the result,
        # and turning the whole thing async from top to bottom would be a pain in the ass.

        return self.current_game.calculate_winner(board)

    
    @called_by(
        GameManager.start_game,
        GameManager.cell_pressed, 
        GameManager.computer_turn_orch
    )
    @on(GameManager.ChangeTurn)         
    @work(exit_on_error=False)
    async def change_turn(self, event: GameManager.ChangeTurn):

        # NOTE: The cell_pressed method in GameManger updates the cell state.
        # That state is reactive, so the cell then updates itself with new rendering.
        # This method is only to update the UI before switching control from human <-> computer.
        # Partially out of necessity because it was more responsive / prevented freezing.

        self.log.debug(f"Turn label changing to {event.value.name}")
        if event.value == PlayerState.PLAYER1:
            self.query_one("#spinner").visible = False
            self.query_one("#turn_label").update("Your turn")
        else:
            self.query_one("#spinner").visible = True
            self.query_one("#turn_label").update("Computer is thinking... ")
            await self.game_manager.computer_turn_orch()

    @called_by(GameManager.end_game)
    @on(GameManager.GameOver)           
    def game_over(self, event: GameManager.GameOver):
        
        self.query_one("#spinner").visible = False
        self.notify("Game over", timeout=1.5)
        if event.result == PlayerState.EMPTY:
            self.query_one("#turn_label").update("It's a tie!")
        else:
            self.query_one("#turn_label").update(f"{event.result.name} wins!")

        self.current_game.grid.clear_focus()            #! This is presuming the game uses a grid.
        self.current_game.query_one("#restart").focus() #! This is presuming there's a button with id "restart"

    @called_by(GameManager.computer_turn_orch)
    @on(GameManager.ComputerMove)
    def play_computer_move(self, event: GameManager.ComputerMove):

        # The AI does not send a cell_pressed event because it doesn't click on the cell,
        # So we need to update the cell state manually.

        cell = self.current_game.grid.query_one(f"#cell_{event.row}_{event.col}")
        cell.state = PlayerState.PLAYER2

    @on(Grid.CellChosen)
    async def cell_chosen(self, event: Grid.CellChosen):
        await self.game_manager.cell_pressed(event.row, event.column)

class GameLoader(DOMNode):

    #* Called by: load_games in TextualGames class.
    def __init__(self, *directories: str):
        super().__init__()
        """ Initialize the game loader with one or more game directories. \n
        Args:
            *directories: Variable number of paths to directories containing games
        """
        self.game_directories = list(directories)
        self.loaded_games: Dict[str, Type[Any]] = {}

    @called_by(TextualGames.load_games)
    @classmethod
    def with_default_games(cls, extra_directories: List[str] = None):
        """ Create a GameLoader that includes both package games and optional extra directories. \n
        ### Args:
            extra_directories: Additional game directories to scan
        ### Returns:
            GameLoader: Configured to load both package and user games
        """
        default_path = games.__path__[0]
        
        directories = [default_path]
        if extra_directories:
            directories.extend(extra_directories)
            
        return cls(*directories)

    @called_by(TextualGames.load_games)
    def discover_games(self) -> Dict[str, Type[Any]]:
        """
        Scan all game directories for .py files and attempt to load them.
        Later directories can override games from earlier directories.
        
        Returns:
            Dict[str, Type[Any]]: Dictionary mapping game names to their game classes
        """
        self.loaded_games.clear()
        
        for directory in self.game_directories:
            if not os.path.exists(directory):
                self.log.error(f"ERROR: Games directory not found: {directory}")
                continue
                
            for filename in os.listdir(directory):
                if filename.endswith('.py') and not filename.startswith('__'):
                    scriptname = filename[:-3]
                    try:
                        module_path = os.path.join(directory, filename)
                        spec = importlib.util.spec_from_file_location(scriptname, module_path)
                        if spec is None or spec.loader is None:
                            self.log.error(f"Failed to load spec for game: {scriptname}")
                            continue
                            
                        module = importlib.util.module_from_spec(spec)
                        spec.loader.exec_module(module)
                        
                        try:
                            game_class = module.loader()
                        except AttributeError:
                            self.log.warning(f"Game {scriptname} does not have an loader function")
                            continue
                        try:
                            game_name = game_class.game_name
                            self.loaded_games[game_name] = game_class
                        except AttributeError:
                            self.log.warning(f"Game {scriptname} does not have a game_name attribute")
                            continue
                            
                    except Exception as e:
                        self.log.error(f"Error loading game {scriptname}: {str(e)}")
                    
        return self.loaded_games




# Textual imports
from rich.spinner import Spinner
from textual.widgets import Static


class SpinnerWidget(Static):
    def __init__(
            self, 
            spinner: str, 
            text: str | None = None, 
            *args, 
            **kwargs
        ):
        super().__init__(*args, **kwargs)
        self._spinner = Spinner(spinner, text)  

    def on_mount(self) -> None:
        self.set_interval(1 / 60, self.update_spinner)

    def update_spinner(self) -> None:
        self.update(self._spinner)


class ScrollingLine(Static):

    def on_mount(self):
        self.counter = 0
        self.max = 3
        self.set_interval(1 / 15, self.asciiscroll)

    def asciiscroll(self):
        asciifoo = (" " * self.counter) + "+" + (" " * ((self.counter - self.max)*-1))
        self.update(asciifoo * (self.screen.size.width//4))
        self.counter += 1
        if self.counter > self.max:
            self.counter = 0
